/*
 * Stateful things in runtime
 */
var Host = {
  mountID: 1,
  component: null,
  isUpdating: false,
  // Roots
  rootComponents: {},
  rootInstances: {},
  // Inject
  driver: null
};

function Element(type, key, ref, props, owner) {
  return {
    // Built-in properties that belong on the element
    type: type,
    key: key,
    ref: ref,
    props: props,
    // Record the component responsible for creating this element.
    _owner: owner
  };
}

function traverseChildren(children, result) {
  if (Array.isArray(children)) {
    for (var i = 0, l = children.length; i < l; i++) {
      traverseChildren(children[i], result);
    }
  } else {
    result.push(children);
  }
}

function flattenChildren(children) {
  if (children == null) {
    return children;
  }

  var result = [];
  traverseChildren(children, result);

  if (result.length === 1) {
    result = result[0];
  }

  return result;
}

var RESERVED_PROPS = {
  key: true,
  ref: true
};

function getRenderErrorInfo() {
  var component = Host.component;

  if (component) {
    var name = component.getName();

    if (name) {
      return ' Check the render method of `' + name + '`.';
    }
  }

  return '';
}

function createElement(type, config, children) {
  if (type == null) {
    throw Error('createElement: type should not be null or undefined.' + getRenderErrorInfo());
  } // Reserved names are extracted


  var props = {};
  var propName;
  var key = null;
  var ref = null;
  var component = Host.component;

  if (config != null) {
    ref = config.ref === undefined ? null : config.ref;
    key = config.key === undefined ? null : String(config.key);

    if (typeof ref === 'string' && !component) {
      console.warn('createElement: adding a string ref outside the component’s render() method.');
    } // Remaining properties are added to a new props object


    for (propName in config) {
      if (!RESERVED_PROPS[propName]) {
        props[propName] = config[propName];
      }
    }
  } // Children arguments can be more than one


  var childrenLength = arguments.length - 2;

  if (childrenLength > 0) {
    if (childrenLength === 1 && !Array.isArray(children)) {
      props.children = children;
    } else {
      var childArray = children;

      if (childrenLength > 1) {
        childArray = new Array(childrenLength);

        for (var i = 0; i < childrenLength; i++) {
          childArray[i] = arguments[i + 2];
        }
      }

      props.children = flattenChildren(childArray);
    }
  } // Resolve default props


  if (type && type.defaultProps) {
    var defaultProps = type.defaultProps;

    for (propName in defaultProps) {
      if (props[propName] === undefined) {
        props[propName] = defaultProps[propName];
      }
    }
  }

  return new Element(type, key, ref, props, component);
}

function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;
  subClass.__proto__ = superClass;
}

/**
 * Base component class.
 */
var Component =
/*#__PURE__*/
function () {
  function Component(props, context, updater) {
    this.props = props;
    this.context = context;
    this.refs = {};
    this.updater = updater;
  }

  var _proto = Component.prototype;

  _proto.isComponentClass = function isComponentClass() {};

  _proto.setState = function setState(partialState, callback) {
    this.updater.setState(this, partialState, callback);
  };

  _proto.forceUpdate = function forceUpdate(callback) {
    this.updater.forceUpdate(this, callback);
  };

  return Component;
}();

var ValueEmitter =
/*#__PURE__*/
function () {
  function ValueEmitter(defaultValue) {
    this.handlers = [];
    this.value = defaultValue;
  }

  var _proto = ValueEmitter.prototype;

  _proto.on = function on(handler) {
    this.handlers.push(handler);
  };

  _proto.off = function off(handler) {
    this.handlers = this.handlers.filter(function (h) {
      return h !== handler;
    });
  };

  _proto.emit = function emit() {
    var _this = this;

    this.handlers.forEach(function (handler) {
      return handler(_this.value);
    });
  };

  return ValueEmitter;
}();

var uniqueId = 0;
function createContext(defaultValue) {
  var _Provider$childContex, _Consumer$contextType;

  var contextProp = '__context_' + uniqueId++ + '__';

  var Provider =
  /*#__PURE__*/
  function (_Component) {
    _inheritsLoose(Provider, _Component);

    function Provider() {
      var _this2;

      _this2 = _Component.call(this) || this;
      _this2.emitter = new ValueEmitter(defaultValue);
      return _this2;
    }

    var _proto2 = Provider.prototype;

    _proto2.getChildContext = function getChildContext() {
      var _ref;

      return _ref = {}, _ref[contextProp] = this.emitter, _ref;
    };

    _proto2.componentWillMount = function componentWillMount() {
      if (this.props.value !== undefined) {
        this.emitter.value = this.props.value;
      }
    };

    _proto2.componentWillReceiveProps = function componentWillReceiveProps(nextProps) {
      if (this.props.value !== nextProps.value) {
        this.emitter.value = nextProps.value;
      }
    };

    _proto2.componentDidUpdate = function componentDidUpdate(prevProps) {
      if (this.props.value !== prevProps.value) {
        this.emitter.emit();
      }
    };

    _proto2.render = function render() {
      return this.props.children;
    };

    return Provider;
  }(Component);

  Provider.childContextTypes = (_Provider$childContex = {}, _Provider$childContex[contextProp] = function () {}, _Provider$childContex);
  Provider.contextProp = contextProp;
  Provider.defaultValue = defaultValue;

  var Consumer =
  /*#__PURE__*/
  function (_Component2) {
    _inheritsLoose(Consumer, _Component2);

    function Consumer(props, context) {
      var _this3;

      _this3 = _Component2.call(this, props, context) || this;
      _this3.state = {
        value: _this3.readContext(_this3.context)
      };

      _this3.onUpdate = function (value) {
        return _this3.state.value !== value && _this3.setState({
          value: value
        });
      };

      return _this3;
    }

    var _proto3 = Consumer.prototype;

    _proto3.readContext = function readContext(context) {
      return context[contextProp] ? context[contextProp].value : defaultValue;
    };

    _proto3.componentDidMount = function componentDidMount() {
      if (this.context[contextProp]) {
        this.context[contextProp].on(this.onUpdate);
      }
    };

    _proto3.componentWillReceiveProps = function componentWillReceiveProps(nextProps, nextContext) {
      var newContextValue = this.readContext(nextContext);

      if (this.state.value !== newContextValue) {
        this.setState({
          value: newContextValue
        });
      }
    };

    _proto3.componentWillUnmount = function componentWillUnmount() {
      if (this.context[contextProp]) {
        this.context[contextProp].off(this.onUpdate);
      }
    };

    _proto3.render = function render() {
      var children = this.props.children;
      var consumer = Array.isArray(children) ? children[0] : children;

      if (typeof consumer === 'function') {
        return consumer(this.state.value);
      }
    };

    return Consumer;
  }(Component);

  Consumer.contextTypes = (_Consumer$contextType = {}, _Consumer$contextType[contextProp] = function () {}, _Consumer$contextType);
  return {
    Provider: Provider,
    Consumer: Consumer
  };
}

function createRef() {
  return {
    current: null
  };
}

var performUpdate = null;

var setImmediatePolyfill = function setImmediatePolyfill(job) {
  return setTimeout(job, 0);
}; // 0s


var scheduleImmediateCallback = typeof setImmediate === 'undefined' ? setImmediatePolyfill : setImmediate;
var cancelImmediateCallback = typeof clearImmediate === 'undefined' ? clearTimeout : clearImmediate;
var beforeNextRenderCallbacks = [];
var beforeNextRenderCallbackId; // Commit before next render

function commit() {
  var preCallbacks = beforeNextRenderCallbacks;
  beforeNextRenderCallbacks = [];
  preCallbacks.forEach(function (callback) {
    return callback();
  });
  preCallbacks = null; // Exec callback maybe schedule a work

  performUpdate();
} // Schedule before next render


var schedule = function schedule(callback) {
  if (beforeNextRenderCallbacks.length === 0) {
    beforeNextRenderCallbackId = scheduleImmediateCallback(commit);
  }

  beforeNextRenderCallbacks.push(callback);
}; // Flush before next render

function flush() {
  if (beforeNextRenderCallbacks.length !== 0) {
    cancelImmediateCallback(beforeNextRenderCallbackId);
    commit();
  }
}
function setUpdater(handle) {
  performUpdate = handle;
}

function getCurrentRenderingInstance() {
  var currentInstance = Host.component._instance;

  if (currentInstance) {
    return currentInstance;
  } else {
    throw Error('Hooks can only be called inside a component.');
  }
}

function areInputsEqual(inputs, prevInputs) {
  for (var i = 0; i < inputs.length; i++) {
    var val1 = inputs[i];
    var val2 = prevInputs[i];

    if (val1 === val2 && (val1 !== 0 || 1 / val1 === 1 / val2) || val1 !== val1 && val2 !== val2 // eslint-disable-line no-self-compare
    ) {
        continue;
      }

    return false;
  }

  return true;
}

function useState(initialState) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    // state lazy initializer
    if (typeof initialState === 'function') {
      initialState = initialState();
    }

    var setState = function setState(newState) {
      var current = hooks[hookID][0];

      if (typeof newState === 'function') {
        newState = newState(current);
      }

      if (newState !== current) {
        // This is a render phase update.  After this render pass, we'll restart
        if (Host.component && Host.component._instance === currentInstance) {
          hooks[hookID][0] = newState;
          currentInstance.isScheduled = true;
        } else {
          !Host.isUpdating && flush();
          hooks[hookID][0] = newState;
          currentInstance.update();
        }
      }
    };

    hooks[hookID] = [initialState, setState];
  }

  return hooks[hookID];
}
function useContext(context) {
  var currentInstance = getCurrentRenderingInstance();
  return currentInstance.readContext(context);
}
function useEffect(effect, inputs) {
  useEffectImpl(effect, inputs, true);
}
function useLayoutEffect(effect, inputs) {
  useEffectImpl(effect, inputs);
}

function useEffectImpl(effect, inputs, defered) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();
  inputs = inputs != null ? inputs : [effect];

  if (!hooks[hookID]) {
    var create = function create(immediately) {
      if (!immediately && defered) return schedule(function () {
        return create(true);
      });
      var current = create.current;

      if (current) {
        // Set this to true to prevent re-entrancy
        var previousIsRendering = Host.isUpdating;
        Host.isUpdating = true;
        destory.current = current();
        create.current = null;
        Host.isUpdating = previousIsRendering;
      }
    };

    var destory = function destory(immediately) {
      if (!immediately && defered) return schedule(function () {
        return destory(true);
      });
      var current = destory.current;

      if (current) {
        // Set this to true to prevent re-entrancy
        var previousIsRendering = Host.isUpdating;
        Host.isUpdating = true;
        current();
        destory.current = null;
        Host.isUpdating = previousIsRendering;
      }
    };

    create.current = effect;
    hooks[hookID] = {
      create: create,
      destory: destory,
      prevInputs: inputs,
      inputs: inputs
    };
    currentInstance.didMount.push(create);
    currentInstance.willUnmount.push(destory);
    currentInstance.didUpdate.push(function () {
      var _hooks$hookID = hooks[hookID],
          prevInputs = _hooks$hookID.prevInputs,
          inputs = _hooks$hookID.inputs,
          create = _hooks$hookID.create;

      if (prevInputs == null || !areInputsEqual(inputs, prevInputs)) {
        destory();
        create();
      }
    });
  } else {
    var hook = hooks[hookID];
    var _create = hook.create,
        prevInputs = hook.inputs;
    hook.inputs = inputs;
    hook.prevInputs = prevInputs;
    _create.current = effect;
  }
}

function useImperativeHandle(ref, create, inputs) {
  var nextInputs = inputs != null ? inputs.concat([ref]) : [ref, create];
  useLayoutEffect(function () {
    if (typeof ref === 'function') {
      ref(create());
      return function () {
        return ref(null);
      };
    } else if (ref != null) {
      ref.current = create();
      return function () {
        ref.current = null;
      };
    }
  }, nextInputs);
}
function useRef(initialValue) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    hooks[hookID] = {
      current: initialValue
    };
  }

  return hooks[hookID];
}
function useCallback(callback, inputs) {
  return useMemo(function () {
    return callback;
  }, inputs);
}
function useMemo(create, inputs) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    hooks[hookID] = [create(), inputs];
  } else {
    var hook = hooks[hookID];
    var prevInputs = hook[1];

    if (!areInputsEqual(inputs, prevInputs)) {
      hook[0] = create();
    }
  }

  return hooks[hookID][0];
}
function useReducer(reducer, initialState, initialAction) {
  var currentInstance = getCurrentRenderingInstance();
  var hookID = currentInstance.getHookID();
  var hooks = currentInstance.getHooks();

  if (!hooks[hookID]) {
    if (initialAction) {
      initialState = reducer(initialState, initialAction);
    }

    var dispatch = function dispatch(action) {
      var hook = hooks[hookID]; // reducer will get in the next render, before that we add all
      // actions to the queue

      var queue = hook[2]; // This is a render phase update.  After this render pass, we'll restart

      if (Host.component && Host.component._instance === currentInstance) {
        queue.push(action);
        currentInstance.isScheduled = true;
      } else {
        !Host.isUpdating && flush();
        queue.push(action);
        currentInstance.update();
      }
    };

    return hooks[hookID] = [initialState, dispatch, []];
  }

  var hook = hooks[hookID];
  var queue = hook[2];
  var next = hook[0];

  for (var i = 0; i < queue.length; i++) {
    next = reducer(next, queue[i]);
  }

  hook[0] = next;
  hook[2] = [];
  return hooks[hookID];
}

var hasOwnProperty = Object.prototype.hasOwnProperty;
/**
 * inlined Object.is polyfill to avoid requiring consumers ship their own
 * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/is
 */

function is(x, y) {
  // SameValue algorithm
  if (x === y) {
    // Steps 1-5, 7-10
    // Steps 6.b-6.e: +0 != -0
    return x !== 0 || 1 / x === 1 / y;
  } else {
    // Step 6.a: NaN == NaN
    return x !== x && y !== y;
  }
}
/**
 * Performs equality by iterating through keys on an object and returning false
 * when any key has values which are not strictly equal between the arguments.
 * Returns true when the values of all keys are strictly equal.
 */


function shallowEqual(objA, objB) {
  if (is(objA, objB)) {
    return true;
  }

  if (typeof objA !== 'object' || objA === null || typeof objB !== 'object' || objB === null) {
    return false;
  }

  var keysA = Object.keys(objA);
  var keysB = Object.keys(objB);

  if (keysA.length !== keysB.length) {
    return false;
  } // Test for A's keys different from B.


  for (var i = 0; i < keysA.length; i++) {
    if (!hasOwnProperty.call(objB, keysA[i]) || !is(objA[keysA[i]], objB[keysA[i]])) {
      return false;
    }
  }

  return true;
}

function memo(type, compare) {
  if (compare === void 0) {
    compare = shallowEqual;
  }

  // Memo could composed
  if (type.compares) {
    type.compares.push(compare);
  } else {
    type.compares = [compare];
  }

  return type;
}

function Fragment(props) {
  return props.children;
}

/**
 * Empty Component
 */

var EmptyComponent =
/*#__PURE__*/
function () {
  function EmptyComponent() {
    this._currentElement = null;
  }

  var _proto = EmptyComponent.prototype;

  _proto.mountComponent = function mountComponent(parent, parentInstance, context, childMounter) {
    this._parent = parent;
    this._parentInstance = parentInstance;
    this._context = context;
    var instance = {
      _internal: this
    };
    var nativeNode = this.getNativeNode();

    if (childMounter) {
      childMounter(nativeNode, parent);
    } else {
      Host.driver.appendChild(nativeNode, parent);
    }

    return instance;
  };

  _proto.unmountComponent = function unmountComponent(notRemoveChild) {
    if (this._nativeNode && !notRemoveChild) {
      Host.driver.removeChild(this._nativeNode, this._parent);
    }

    this._nativeNode = null;
    this._parent = null;
    this._parentInstance = null;
    this._context = null;
  };

  _proto.updateComponent = function updateComponent() {// Noop
  };

  _proto.getNativeNode = function getNativeNode() {
    // Weex native node
    if (this._nativeNode == null) {
      this._nativeNode = Host.driver.createEmpty();
    }

    return this._nativeNode;
  };

  return EmptyComponent;
}();

/*
 * Ref manager
 */
var Ref = {
  update: function update(prevElement, nextElement, component) {
    var prevRef = prevElement && prevElement.ref || null;
    var nextRef = nextElement && nextElement.ref || null; // Update refs in owner component

    if (prevRef !== nextRef) {
      // Detach prev RenderedElement's ref
      prevRef != null && this.detach(prevElement._owner, prevRef, component); // Attach next RenderedElement's ref

      nextRef != null && this.attach(nextElement._owner, nextRef, component);
    }
  },
  attach: function attach(ownerComponent, ref, component) {
    if (!ownerComponent) {
      throw Error('Multiple version of Rax used in project.');
    }

    var instance = component.getPublicInstance();

    if (typeof ref === 'function') {
      ref(instance);
    } else if (typeof ref === 'object') {
      ref.current = instance;
    } else {
      ownerComponent._instance.refs[ref] = instance;
    }
  },
  detach: function detach(ownerComponent, ref, component) {
    if (typeof ref === 'function') {
      // When the referenced component is unmounted and whenever the ref changes, the old ref will be called with null as an argument.
      ref(null);
    } else {
      // Must match component and ref could detach the ref on owner when A's before ref is B's current ref
      var instance = component.getPublicInstance();

      if (typeof ref === 'object' && ref.current === instance) {
        ref.current = null;
      } else if (ownerComponent._instance.refs[ref] === instance) {
        delete ownerComponent._instance.refs[ref];
      }
    }
  }
};

function instantiateComponent(element) {
  var instance;

  if (element === undefined || element === null || element === false || element === true) {
    instance = new Host.Empty();
  } else if (Array.isArray(element)) {
    instance = new Host.Fragment(element);
  } else if (typeof element === 'object' && element.type) {
    // Special case string values
    if (typeof element.type === 'string') {
      instance = new Host.Native(element);
    } else {
      instance = new Host.Composite(element);
    }
  } else if (typeof element === 'string' || typeof element === 'number') {
    instance = new Host.Text(element);
  } else {
    throwInvalidComponentError(element);
  }

  instance._mountIndex = 0;
  return instance;
}

function throwInvalidComponentError(element) {
  throw Error("Invalid element type: " + element + ". (current: " + (typeof element === 'object' && Object.keys(element) || typeof element) + ")");
}

function shouldUpdateComponent(prevElement, nextElement) {
  // TODO: prevElement and nextElement could be array
  var prevEmpty = prevElement === null;
  var nextEmpty = nextElement === null;

  if (prevEmpty || nextEmpty) {
    return prevEmpty === nextEmpty;
  }

  var prevType = typeof prevElement;
  var nextType = typeof nextElement;

  if (prevType === 'string' || prevType === 'number') {
    return nextType === 'string' || nextType === 'number';
  } else {
    return prevType === 'object' && nextType === 'object' && prevElement.type === nextElement.type && prevElement.key === nextElement.key;
  }
}

var getElementKeyName = (function (children, element, index) {
  var elementKey = element && element.key;
  var defaultName = '.' + index.toString(36); // Inner child name default format fallback
  // Key should must be string type

  if (typeof elementKey === 'string') {
    var keyName = '$' + elementKey; // Child keys must be unique.

    var keyUnique = children[keyName] === undefined; // Only the first child will be used when encountered two children with the same key

    if (!keyUnique) console.warn("Encountered two children with the same key \"" + elementKey + "\".");
    return keyUnique ? keyName : defaultName;
  } else {
    return defaultName;
  }
});

var rootID = 1;

var Root =
/*#__PURE__*/
function (_Component) {
  _inheritsLoose(Root, _Component);

  function Root() {
    var _this;

    _this = _Component.call(this) || this;
    _this.state = {
      // Using fragment instead of null for avoid create a comment node when init mount
      element: []
    };
    _this.rootID = rootID++;
    return _this;
  }

  var _proto = Root.prototype;

  _proto.getPublicInstance = function getPublicInstance() {
    return this.getRenderedComponent().getPublicInstance();
  };

  _proto.getRenderedComponent = function getRenderedComponent() {
    return this._internal._renderedComponent;
  };

  _proto.update = function update(element) {
    this.setState({
      element: element
    });
  };

  _proto.render = function render() {
    return this.state.element;
  };

  return Root;
}(Component);

/**
 * Instance manager
 */

var KEY = '__r';
var Instance = {
  set: function set(node, instance) {
    if (!node[KEY]) {
      node[KEY] = instance; // Record root instance to roots map

      if (instance.rootID) {
        Host.rootInstances[instance.rootID] = instance;
        Host.rootComponents[instance.rootID] = instance._internal;
      }
    }
  },
  get: function get(node) {
    return node[KEY];
  },
  remove: function remove(node) {
    var instance = this.get(node);

    if (instance) {
      node[KEY] = null;

      if (instance.rootID) {
        delete Host.rootComponents[instance.rootID];
        delete Host.rootInstances[instance.rootID];
      }
    }
  },
  mount: function mount(element, container, parentInstance) {

    var driver = Host.driver; // Before render callback

    driver.beforeRender && driver.beforeRender(); // Real native root node is body

    if (container == null) {
      container = driver.createBody();
    } // Get the context from the conceptual parent component.


    var parentContext;

    if (parentInstance) {
      var parentInternal = parentInstance._internal;
      parentContext = parentInternal._processChildContext(parentInternal._context);
    } // Update root component


    var prevRootInstance = this.get(container);

    if (prevRootInstance && prevRootInstance.rootID) {
      if (parentContext) {
        // Using _penddingContext to pass new context
        prevRootInstance._internal._penddingContext = parentContext;
      }

      prevRootInstance.update(element);
      return prevRootInstance;
    } // Init root component with empty children


    var renderedComponent = instantiateComponent(createElement(Root));
    var defaultContext = parentContext || {};
    var rootInstance = renderedComponent.mountComponent(container, null, defaultContext);
    this.set(container, rootInstance); // Mount new element through update queue avoid when there is in rendering phase

    rootInstance.update(element); // After render callback

    driver.afterRender && driver.afterRender(rootInstance);

    return rootInstance;
  }
};

var STYLE = 'style';
var CHILDREN = 'children';
var TREE = 'tree';
var EVENT_PREFIX_REGEXP = /^on[A-Z]/;
/**
 * Native Component
 */

var NativeComponent =
/*#__PURE__*/
function () {
  function NativeComponent(element) {
    this._currentElement = element;
  }

  var _proto = NativeComponent.prototype;

  _proto.mountComponent = function mountComponent(parent, parentInstance, context, childMounter) {
    // Parent native element
    this._parent = parent;
    this._parentInstance = parentInstance;
    this._context = context;
    this._mountID = Host.mountID++;
    var props = this._currentElement.props;
    var type = this._currentElement.type;
    var instance = {
      _internal: this,
      type: type,
      props: props
    };
    var appendType = props.append || TREE; // Default is node

    this._instance = instance; // Clone a copy for style diff

    this._prevStyleCopy = Object.assign({}, props.style);
    var nativeNode = this.getNativeNode();

    if (appendType !== TREE) {
      if (childMounter) {
        childMounter(nativeNode, parent);
      } else {
        Host.driver.appendChild(nativeNode, parent);
      }
    }

    if (this._currentElement && this._currentElement.ref) {
      Ref.attach(this._currentElement._owner, this._currentElement.ref, this);
    } // Process children


    var children = props.children;

    if (children != null) {
      this.mountChildren(children, context);
    }

    if (appendType === TREE) {
      if (childMounter) {
        childMounter(nativeNode, parent);
      } else {
        Host.driver.appendChild(nativeNode, parent);
      }
    }

    return instance;
  };

  _proto.mountChildren = function mountChildren(children, context) {
    var _this = this;

    if (!Array.isArray(children)) {
      children = [children];
    }

    var renderedChildren = this._renderedChildren = {};
    var nativeNode = this.getNativeNode();
    var renderedChildrenImage = children.map(function (element, index) {
      var renderedChild = instantiateComponent(element);
      var name = getElementKeyName(renderedChildren, element, index);
      renderedChildren[name] = renderedChild;
      renderedChild._mountIndex = index; // Mount

      var mountImage = renderedChild.mountComponent(nativeNode, _this._instance, context);
      return mountImage;
    });
    return renderedChildrenImage;
  };

  _proto.unmountChildren = function unmountChildren(notRemoveChild) {
    var renderedChildren = this._renderedChildren;

    if (renderedChildren) {
      for (var name in renderedChildren) {
        var renderedChild = renderedChildren[name];
        renderedChild.unmountComponent(notRemoveChild);
      }

      this._renderedChildren = null;
    }
  };

  _proto.unmountComponent = function unmountComponent(notRemoveChild) {
    if (this._nativeNode) {
      var ref = this._currentElement.ref;

      if (ref) {
        Ref.detach(this._currentElement._owner, ref, this);
      }

      Instance.remove(this._nativeNode);

      if (!notRemoveChild) {
        Host.driver.removeChild(this._nativeNode, this._parent);
      }
    }

    this.unmountChildren(notRemoveChild);

    this._currentElement = null;
    this._nativeNode = null;
    this._parent = null;
    this._parentInstance = null;
    this._context = null;
    this._instance = null;
    this._prevStyleCopy = null;
  };

  _proto.updateComponent = function updateComponent(prevElement, nextElement, prevContext, nextContext) {
    // Replace current element
    this._currentElement = nextElement;
    Ref.update(prevElement, nextElement, this);
    var prevProps = prevElement.props;
    var nextProps = nextElement.props;
    this.updateProperties(prevProps, nextProps);
    this.updateChildren(nextProps.children, nextContext);
  };

  _proto.updateProperties = function updateProperties(prevProps, nextProps) {
    var propKey;
    var styleName;
    var styleUpdates;
    var driver = Host.driver;
    var nativeNode = this.getNativeNode();

    for (propKey in prevProps) {
      if (propKey === CHILDREN || nextProps.hasOwnProperty(propKey) || !prevProps.hasOwnProperty(propKey) || prevProps[propKey] == null) {
        continue;
      }

      if (propKey === STYLE) {
        var lastStyle = this._prevStyleCopy;

        for (styleName in lastStyle) {
          if (lastStyle.hasOwnProperty(styleName)) {
            styleUpdates = styleUpdates || {};
            styleUpdates[styleName] = '';
          }
        }

        this._prevStyleCopy = null;
      } else if (EVENT_PREFIX_REGEXP.test(propKey)) {
        if (typeof prevProps[propKey] === 'function') {
          driver.removeEventListener(nativeNode, propKey.slice(2).toLowerCase(), prevProps[propKey]);
        }
      } else {
        driver.removeAttribute(nativeNode, propKey, prevProps[propKey]);
      }
    }

    for (propKey in nextProps) {
      var nextProp = nextProps[propKey];
      var prevProp = propKey === STYLE ? this._prevStyleCopy : prevProps != null ? prevProps[propKey] : undefined;

      if (propKey === CHILDREN || !nextProps.hasOwnProperty(propKey) || nextProp === prevProp || nextProp == null && prevProp == null) {
        continue;
      } // Update style


      if (propKey === STYLE) {
        if (nextProp) {
          // Clone property
          nextProp = this._prevStyleCopy = Object.assign({}, nextProp);
        } else {
          this._prevStyleCopy = null;
        }

        if (prevProp != null) {
          // Unset styles on `prevProp` but not on `nextProp`.
          for (styleName in prevProp) {
            if (prevProp.hasOwnProperty(styleName) && (!nextProp || !nextProp.hasOwnProperty(styleName))) {
              styleUpdates = styleUpdates || {};
              styleUpdates[styleName] = '';
            }
          } // Update styles that changed since `prevProp`.


          for (styleName in nextProp) {
            if (nextProp.hasOwnProperty(styleName) && prevProp[styleName] !== nextProp[styleName]) {
              styleUpdates = styleUpdates || {};
              styleUpdates[styleName] = nextProp[styleName];
            }
          }
        } else {
          // Assign next prop when prev style is null
          styleUpdates = nextProp;
        }
      } else if (EVENT_PREFIX_REGEXP.test(propKey)) {
        // Update event binding
        var eventName = propKey.slice(2).toLowerCase();

        if (typeof prevProp === 'function') {
          driver.removeEventListener(nativeNode, eventName, prevProp, nextProps);
        }

        if (typeof nextProp === 'function') {
          driver.addEventListener(nativeNode, eventName, nextProp, nextProps);
        }
      } else {
        // Update other property
        var payload = {};
        payload[propKey] = nextProp;

        if (nextProp != null) {
          driver.setAttribute(nativeNode, propKey, nextProp);
        } else {
          driver.removeAttribute(nativeNode, propKey, prevProps[propKey]);
        }
      }
    }

    if (styleUpdates) {

      driver.setStyles(nativeNode, styleUpdates);
    }
  };

  _proto.updateChildren = function updateChildren(nextChildrenElements, context) {
    // prev rendered children
    var prevChildren = this._renderedChildren;
    var driver = Host.driver;

    if (nextChildrenElements == null && prevChildren == null) {
      return;
    }

    var nextChildren = {};

    if (nextChildrenElements != null) {
      if (!Array.isArray(nextChildrenElements)) {
        nextChildrenElements = [nextChildrenElements];
      } // Update next children elements


      for (var index = 0, length = nextChildrenElements.length; index < length; index++) {
        var nextElement = nextChildrenElements[index];
        var name = getElementKeyName(nextChildren, nextElement, index);
        var prevChild = prevChildren && prevChildren[name];
        var prevElement = prevChild && prevChild._currentElement;
        var prevContext = prevChild && prevChild._context; // Try to update between the two of some name that has some element type,
        // and move child in next children loop if need

        if (prevChild != null && shouldUpdateComponent(prevElement, nextElement)) {
          if (prevElement !== nextElement || prevContext !== context) {
            // Pass the same context when updating chidren
            prevChild.updateComponent(prevElement, nextElement, context, context);
          }

          nextChildren[name] = prevChild;
        } else {
          // Unmount the prevChild when some name with nextChild but different element type,
          // and move child node in next children loop
          if (prevChild) {
            prevChild._unmount = true;
          } // The child must be instantiated before it's mounted.


          nextChildren[name] = instantiateComponent(nextElement);
        }
      }
    }

    var prevFirstChild;
    var prevFirstNativeNode;
    var shouldUnmountPrevFirstChild; // Unmount children that are no longer present.

    if (prevChildren != null) {
      for (var _name in prevChildren) {
        if (!prevChildren.hasOwnProperty(_name)) {
          continue;
        }

        var _prevChild = prevChildren[_name];
        var shouldUnmount = _prevChild._unmount || !nextChildren[_name]; // Store old first child ref for append node ahead and maybe delay remove it

        if (!prevFirstChild) {
          shouldUnmountPrevFirstChild = shouldUnmount;
          prevFirstChild = _prevChild;
          prevFirstNativeNode = prevFirstChild.getNativeNode();

          if (Array.isArray(prevFirstNativeNode)) {
            prevFirstNativeNode = prevFirstNativeNode[0];
          }
        } else if (shouldUnmount) {
          _prevChild.unmountComponent();
        }
      }
    }

    if (nextChildren != null) {
      var insertNodes = function insertNodes(nativeNodes, parent) {
        // The nativeNodes maybe fragment, so convert to array type
        if (!Array.isArray(nativeNodes)) {
          nativeNodes = [nativeNodes];
        }

        if (lastPlacedNode) {
          // Should reverse order when insert new child after lastPlacedNode:
          // [lastPlacedNode, *newChild1, *newChild2]
          for (var i = nativeNodes.length - 1; i >= 0; i--) {
            driver.insertAfter(nativeNodes[i], lastPlacedNode);
          }
        } else if (prevFirstNativeNode) {
          // [*newChild1, *newChild2, prevFirstNativeNode]
          for (var _i = 0; _i < nativeNodes.length; _i++) {
            driver.insertBefore(nativeNodes[_i], prevFirstNativeNode);
          }
        } else if (parent) {
          // [*newChild1, *newChild2]
          for (var _i2 = 0; _i2 < nativeNodes.length; _i2++) {
            driver.appendChild(nativeNodes[_i2], parent);
          }
        }
      };

      // `nextIndex` will increment for each child in `nextChildren`
      var nextIndex = 0;
      var lastPlacedNode = null;
      var nextNativeNode = [];

      for (var _name2 in nextChildren) {
        if (!nextChildren.hasOwnProperty(_name2)) {
          continue;
        }

        var nextChild = nextChildren[_name2];

        var _prevChild2 = prevChildren && prevChildren[_name2]; // Try to move the some key prevChild but current not at the some position


        if (_prevChild2 === nextChild) {
          var prevChildNativeNode = _prevChild2.getNativeNode();

          if (_prevChild2._mountIndex !== nextIndex) {
            insertNodes(prevChildNativeNode);
          }
        } else {
          // Mount nextChild that in prevChildren there has no some name
          var parent = this.getNativeNode(); // Fragment extended native component, so if parent is fragment should get this._parent

          if (Array.isArray(parent)) {
            parent = this._parent;
          }

          nextChild.mountComponent(parent, this._instance, context, insertNodes // Insert nodes mounter
          );
        } // Update to the latest mount order


        nextChild._mountIndex = nextIndex++; // Get the last child

        lastPlacedNode = nextChild.getNativeNode(); // Push to nextNativeNode

        nextNativeNode = nextNativeNode.concat(lastPlacedNode);

        if (Array.isArray(lastPlacedNode)) {
          lastPlacedNode = lastPlacedNode[lastPlacedNode.length - 1];
        }
      } // Sync update native refs


      if (Array.isArray(this._nativeNode)) {
        // Clear all and push the new array
        this._nativeNode.splice(0, this._nativeNode.length);

        for (var i = 0; i < nextNativeNode.length; i++) {
          this._nativeNode.push(nextNativeNode[i]);
        }
      }
    }

    if (shouldUnmountPrevFirstChild) {
      prevFirstChild.unmountComponent();
    }

    this._renderedChildren = nextChildren;
  };

  _proto.getNativeNode = function getNativeNode() {
    if (this._nativeNode == null) {
      this._nativeNode = Host.driver.createElement(this._instance);
      Instance.set(this._nativeNode, this._instance);
    }

    return this._nativeNode;
  };

  _proto.getPublicInstance = function getPublicInstance() {
    return this.getNativeNode();
  };

  _proto.getName = function getName() {
    return this._currentElement.type;
  };

  return NativeComponent;
}();

/**
 * Text Component
 */

var TextComponent =
/*#__PURE__*/
function () {
  function TextComponent(element) {
    this._currentElement = element; // Text string

    this._text = String(element);
  }

  var _proto = TextComponent.prototype;

  _proto.mountComponent = function mountComponent(parent, parentInstance, context, childMounter) {
    this._parent = parent;
    this._parentInstance = parentInstance;
    this._context = context;
    this._mountID = Host.mountID++; // Weex dom operation

    var nativeNode = this.getNativeNode();

    if (childMounter) {
      childMounter(nativeNode, parent);
    } else {
      Host.driver.appendChild(nativeNode, parent);
    }

    var instance = {
      _internal: this
    };

    return instance;
  };

  _proto.unmountComponent = function unmountComponent(notRemoveChild) {
    if (this._nativeNode && !notRemoveChild) {
      Host.driver.removeChild(this._nativeNode, this._parent);
    }

    this._currentElement = null;
    this._nativeNode = null;
    this._parent = null;
    this._parentInstance = null;
    this._context = null;
    this._text = null;
  };

  _proto.updateComponent = function updateComponent(prevElement, nextElement, context) {
    // If some text do noting
    if (prevElement !== nextElement) {
      // Replace current element
      this._currentElement = nextElement; // Devtool read the latest stringText value

      this._text = String(nextElement);
      Host.driver.updateText(this.getNativeNode(), this._text);
    }
  };

  _proto.getNativeNode = function getNativeNode() {
    if (this._nativeNode == null) {
      this._nativeNode = Host.driver.createText(this._text);
    }

    return this._nativeNode;
  };

  return TextComponent;
}();

var RE_RENDER_LIMIT = 24;
/**
 * Functional Reactive Component Class Wrapper
 */

var ReactiveComponent =
/*#__PURE__*/
function (_Component) {
  _inheritsLoose(ReactiveComponent, _Component);

  function ReactiveComponent(pureRender, ref) {
    var _this;

    _this = _Component.call(this) || this; // A pure function

    _this._render = pureRender;
    _this._hookID = 0; // Number of rerenders

    _this._reRenders = 0;
    _this._hooks = {}; // Handles store

    _this.didMount = [];
    _this.didUpdate = [];
    _this.willUnmount = []; // Is render scheduled

    _this.isScheduled = false;

    if (pureRender.forwardRef) {
      _this.prevForwardRef = _this.forwardRef = ref;
    }

    var compares = pureRender.compares;

    if (compares) {
      _this.shouldComponentUpdate = function (nextProps) {
        // Process composed compare
        var arePropsEqual = true; // Compare push in and pop out

        for (var i = compares.length - 1; i > -1; i--) {
          if (arePropsEqual = compares[i](_this.props, nextProps)) {
            break;
          }
        }

        return !arePropsEqual || _this.prevForwardRef !== _this.forwardRef;
      };
    }

    return _this;
  }

  var _proto = ReactiveComponent.prototype;

  _proto.getHooks = function getHooks() {
    return this._hooks;
  };

  _proto.getHookID = function getHookID() {
    return ++this._hookID;
  };

  _proto.readContext = function readContext(context) {
    var _this2 = this;

    var Provider = context.Provider;
    var unmaskContext = this._internal._context;
    var contextProp = Provider.contextProp;
    var contextEmitter = unmaskContext[contextProp];

    if (contextEmitter) {
      var mountId = this._internal._mountID;

      if (!contextEmitter[mountId]) {
        // One context one updater bind
        contextEmitter[mountId] = {};

        var contextUpdater = function contextUpdater(newContext) {
          if (newContext !== contextEmitter[mountId].renderedContext) {
            _this2.update();
          }
        };

        contextEmitter.on(contextUpdater);
        this.willUnmount.push(function () {
          delete contextEmitter[mountId];
          contextEmitter.off(contextUpdater);
        });
      }

      return contextEmitter[mountId].renderedContext = contextEmitter.value;
    }

    return Provider.defaultValue;
  };

  _proto.componentDidMount = function componentDidMount() {
    this.didMount.forEach(function (handler) {
      return handler();
    });
  };

  _proto.componentDidUpdate = function componentDidUpdate() {
    this.didUpdate.forEach(function (handler) {
      return handler();
    });
  };

  _proto.componentWillUnmount = function componentWillUnmount() {
    this.willUnmount.forEach(function (handler) {
      return handler();
    });
  };

  _proto.update = function update() {
    this.forceUpdate();
  };

  _proto.render = function render() {

    this._hookID = 0;
    this._reRenders = 0;
    this.isScheduled = false;

    var children = this._render(this.props, this.forwardRef ? this.forwardRef : this.context);

    while (this.isScheduled) {
      this._reRenders++;

      if (this._reRenders > RE_RENDER_LIMIT) {
        throw Error('Too many re-renders, the number of renders is limited to prevent an infinite loop.');
      }

      this._hookID = 0;
      this.isScheduled = false;
      children = this._render(this.props, this.forwardRef ? this.forwardRef : this.context);
    }

    return children;
  };

  return ReactiveComponent;
}(Component);

var dirtyComponents = [];

function enqueueCallback(internal, callback) {
  var callbackQueue = internal._pendingCallbacks || (internal._pendingCallbacks = []);
  callbackQueue.push(callback);
}

function enqueueState(internal, partialState) {
  var stateQueue = internal._pendingStateQueue || (internal._pendingStateQueue = []);
  stateQueue.push(partialState);
}

function runCallbacks(callbacks, context) {
  if (callbacks) {
    for (var i = 0; i < callbacks.length; i++) {
      callbacks[i].call(context);
    }
  }
}

function runUpdate(component) {
  var internal = component._internal;

  if (!internal) {
    return;
  }

  Host.isUpdating = true; // If updateComponent happens to enqueue any new updates, we
  // shouldn't execute the callbacks until the next render happens, so
  // stash the callbacks first

  var callbacks = internal._pendingCallbacks;
  internal._pendingCallbacks = null;
  var prevElement = internal._currentElement;
  var prevUnmaskedContext = internal._context;
  var nextUnmaskedContext = internal._penddingContext || prevUnmaskedContext;
  internal._penddingContext = undefined;

  if (internal._pendingStateQueue || internal._pendingForceUpdate) {
    internal.updateComponent(prevElement, prevElement, prevUnmaskedContext, nextUnmaskedContext);
  }

  runCallbacks(callbacks, component);
  Host.isUpdating = false;
}

function mountOrderComparator(c1, c2) {
  return c2._internal._mountID - c1._internal._mountID;
}

function performUpdate$1() {
  if (Host.isUpdating) {
    return;
  }

  var dirties;
  var component;

  while (dirtyComponents.length > 0) {
    // Before next render, we will flush all the PassiveEffects
    flush(); // Since reconciling a component higher in the owner hierarchy usually (not
    // always -- see shouldComponentUpdate()) will reconcile children, reconcile
    // them before their children by sorting the array.

    dirties = dirtyComponents.sort(mountOrderComparator); // Reset dirtyComponents, using to collect dirtyComponent
    // generated by the next render

    dirtyComponents = [];

    while (component = dirties.pop()) {
      runUpdate(component);
    }
  }
}

function scheduleUpdate(component) {
  if (dirtyComponents.indexOf(component) < 0) {
    dirtyComponents.push(component);
  }

  performUpdate$1();
}

function requestUpdate(component, partialState, callback) {
  var internal = component._internal;

  if (!internal) {
    return;
  }

  !Host.isUpdating && flush();

  if (callback) {
    enqueueCallback(internal, callback);
  }

  var hasComponentRendered = internal._renderedComponent; // setState

  if (partialState) {
    enqueueState(internal, partialState); // State pending in componentWillReceiveProps and componentWillMount

    if (!internal._pendingState && hasComponentRendered) {
      scheduleUpdate(component);
    }
  } else {
    // forceUpdate
    internal._pendingForceUpdate = true;

    if (hasComponentRendered) {
      scheduleUpdate(component);
    }
  }
}

setUpdater(performUpdate$1);
var Updater = {
  setState: requestUpdate,
  forceUpdate: function forceUpdate(component, callback) {
    requestUpdate(component, null, callback);
  },
  runCallbacks: runCallbacks
};

function performInSandbox(fn, instance, callback) {
  try {
    return fn();
  } catch (e) {
    if (callback) {
      callback(e);
    } else {
      handleError(instance, e);
    }
  }
}

function handleError(instance, error) {
  var boundary;

  while (instance) {
    var internal = instance._internal;

    if (typeof instance.componentDidCatch === 'function') {
      boundary = instance;
      break;
    } else if (internal && internal._parentInstance) {
      instance = internal._parentInstance;
    } else {
      break;
    }
  }

  if (boundary) {
    // should not attempt to recover an unmounting error boundary
    var boundaryInternal = boundary._internal;

    if (boundaryInternal) {
      var callbackQueue = boundaryInternal._pendingCallbacks || (boundaryInternal._pendingCallbacks = []);
      callbackQueue.push(function () {
        return boundary.componentDidCatch(error);
      });
    }
  } else {
    // Do not break when error happens
    setTimeout(function () {
      throw error;
    }, 0);
  }
}
/**
 * Composite Component
 */


var CompositeComponent =
/*#__PURE__*/
function () {
  function CompositeComponent(element) {
    this._currentElement = element;
  }

  var _proto = CompositeComponent.prototype;

  _proto.getName = function getName() {
    var type = this._currentElement.type;
    var instance = this._instance;
    var constructor = instance && instance.constructor;
    return type.displayName || constructor && constructor.displayName || type.name || constructor && constructor.name || null;
  };

  _proto.mountComponent = function mountComponent(parent, parentInstance, context, childMounter) {

    this._parent = parent;
    this._parentInstance = parentInstance;
    this._context = context;
    this._mountID = Host.mountID++;
    this._updateCount = 0;

    var currentElement = this._currentElement;
    var Component = currentElement.type;
    var ref = currentElement.ref;
    var publicProps = currentElement.props;
    var isClass = Component.prototype;
    var isComponentClass = isClass && Component.prototype.isComponentClass; // Class stateless component without state but have lifecycles

    var isStatelessClass = isClass && Component.prototype.render; // Context process

    var publicContext = this._processContext(context); // Initialize the public class


    var instance;
    var renderedElement;

    try {
      if (isComponentClass || isStatelessClass) {
        // Component instance
        instance = new Component(publicProps, publicContext, Updater);
      } else if (typeof Component === 'function') {
        // Functional reactive component with hooks
        instance = new ReactiveComponent(Component, ref);
      } else {
        throwInvalidComponentError(Component);
      }
    } catch (e) {
      handleError(parentInstance, e);
      return instance;
    } // These should be set up in the constructor, but as a convenience for
    // simpler class abstractions, we set them up after the fact.


    instance.props = publicProps;
    instance.context = publicContext;
    instance.refs = {}; // Inject the updater into instance

    instance.updater = Updater;
    instance._internal = this;
    this._instance = instance; // Init state, must be set to an object or null

    var initialState = instance.state;

    if (initialState === undefined) {
      // TODO clone the state?
      instance.state = initialState = null;
    }

    var error = null;

    var errorCallback = function errorCallback(e) {
      error = e;
    };

    if (instance.componentWillMount) {
      performInSandbox(function () {
        {
          instance.componentWillMount();
        }
      }, instance, errorCallback);
    }

    if (renderedElement == null) {
      Host.component = this; // Process pending state when call setState in componentWillMount

      instance.state = this._processPendingState(publicProps, publicContext);
      performInSandbox(function () {
        {
          renderedElement = instance.render();
        }
      }, instance, errorCallback);
      Host.component = null;
    }

    this._renderedComponent = instantiateComponent(renderedElement);

    this._renderedComponent.mountComponent(this._parent, instance, this._processChildContext(context), childMounter);

    if (error) {
      handleError(instance, error);
    }

    if (!currentElement.type.forwardRef && ref) {
      Ref.attach(currentElement._owner, ref, this);
    }

    if (instance.componentDidMount) {
      performInSandbox(function () {
        {
          instance.componentDidMount();
        }
      }, instance);
    } // Trigger setState callback in componentWillMount or boundary callback after rendered


    var callbacks = this._pendingCallbacks;

    if (callbacks) {
      this._pendingCallbacks = null;
      Updater.runCallbacks(callbacks, instance);
    }

    return instance;
  };

  _proto.unmountComponent = function unmountComponent(notRemoveChild) {
    var instance = this._instance;

    if (!instance) {
      return;
    }

    if (instance.componentWillUnmount) {
      performInSandbox(function () {
        instance.componentWillUnmount();
      }, instance);
    }

    instance._internal = null;

    if (this._renderedComponent != null) {
      var currentElement = this._currentElement;
      var ref = currentElement.ref;

      if (!currentElement.type.forwardRef && ref) {
        Ref.detach(currentElement._owner, ref, this);
      }

      this._renderedComponent.unmountComponent(notRemoveChild);

      this._renderedComponent = null;
      this._instance = null;
    }

    this._currentElement = null;
    this._parentInstance = null; // Reset pending fields
    // Even if this component is scheduled for another update in ReactUpdates,
    // it would still be ignored because these fields are reset.

    this._pendingStateQueue = null;
    this._pendingForceUpdate = false; // These fields do not really need to be reset since this object is no
    // longer accessible.

    this._context = null;
  }
  /**
   * Filters the context object to only contain keys specified in
   * `contextTypes`
   */
  ;

  _proto._processContext = function _processContext(context) {
    var Component = this._currentElement.type;
    var contextTypes = Component.contextTypes;

    if (!contextTypes) {
      return {};
    }

    var maskedContext = {};

    for (var contextName in contextTypes) {
      maskedContext[contextName] = context[contextName];
    }

    return maskedContext;
  };

  _proto._processChildContext = function _processChildContext(currentContext) {
    var instance = this._instance;
    var childContext = instance.getChildContext && instance.getChildContext();

    if (childContext) {
      return Object.assign({}, currentContext, childContext);
    }

    return currentContext;
  };

  _proto._processPendingState = function _processPendingState(props, context) {
    var instance = this._instance;
    var queue = this._pendingStateQueue;

    if (!queue) {
      return instance.state;
    } // Reset pending queue


    this._pendingStateQueue = null;
    var nextState = Object.assign({}, instance.state);

    for (var i = 0; i < queue.length; i++) {
      var partial = queue[i];
      Object.assign(nextState, typeof partial === 'function' ? partial.call(instance, nextState, props, context) : partial);
    }

    return nextState;
  };

  _proto.updateComponent = function updateComponent(prevElement, nextElement, prevUnmaskedContext, nextUnmaskedContext) {
    var instance = this._instance;

    if (!instance) {
      console.error("Update component '" + this.getName() + "' that has already been unmounted (or failed to mount).");
    }

    var willReceive = false;
    var nextContext;
    var nextProps; // Determine if the context has changed or not

    if (this._context === nextUnmaskedContext) {
      nextContext = instance.context;
    } else {
      nextContext = this._processContext(nextUnmaskedContext);
      willReceive = true;
    } // Distinguish between a props update versus a simple state update


    if (prevElement === nextElement) {
      // Skip checking prop types again -- we don't read component.props to avoid
      // warning for DOM component props in this upgrade
      nextProps = nextElement.props;
    } else {
      nextProps = nextElement.props;
      willReceive = true;
    }

    var hasReceived = willReceive && instance.componentWillReceiveProps;

    if (hasReceived) {
      // Calling this.setState() within componentWillReceiveProps will not trigger an additional render.
      this._pendingState = true;
      performInSandbox(function () {
        instance.componentWillReceiveProps(nextProps, nextContext);
      }, instance);
      this._pendingState = false;
    } // Update refs


    if (this._currentElement.type.forwardRef) {
      instance.prevForwardRef = prevElement.ref;
      instance.forwardRef = nextElement.ref;
    } else {
      Ref.update(prevElement, nextElement, this);
    } // Shoud update always default


    var shouldUpdate = true;
    var prevProps = instance.props;
    var prevState = instance.state; // TODO: could delay execution processPendingState

    var nextState = this._processPendingState(nextProps, nextContext); // ShouldComponentUpdate is not called when forceUpdate is used


    if (!this._pendingForceUpdate) {
      if (instance.shouldComponentUpdate) {
        shouldUpdate = performInSandbox(function () {
          return instance.shouldComponentUpdate(nextProps, nextState, nextContext);
        }, instance);
      } else if (instance.isPureComponentClass) {
        shouldUpdate = !shallowEqual(prevProps, nextProps) || !shallowEqual(prevState, nextState);
      }
    }

    if (shouldUpdate) {
      this._pendingForceUpdate = false; // Will set `this.props`, `this.state` and `this.context`.

      var prevContext = instance.context; // Cannot use this.setState() in componentWillUpdate.
      // If need to update state in response to a prop change, use componentWillReceiveProps instead.

      performInSandbox(function () {
        if (instance.componentWillUpdate) {
          instance.componentWillUpdate(nextProps, nextState, nextContext);
        }
      }, instance); // Replace with next

      this._currentElement = nextElement;
      this._context = nextUnmaskedContext;
      instance.props = nextProps;
      instance.state = nextState;
      instance.context = nextContext;

      this._updateRenderedComponent(nextUnmaskedContext);

      performInSandbox(function () {
        if (instance.componentDidUpdate) {
          instance.componentDidUpdate(prevProps, prevState, prevContext);
        }
      }, instance);
      this._updateCount++;
    } else {
      // If it's determined that a component should not update, we still want
      // to set props and state but we shortcut the rest of the update.
      this._currentElement = nextElement;
      this._context = nextUnmaskedContext;
      instance.props = nextProps;
      instance.state = nextState;
      instance.context = nextContext;
    } // Flush setState callbacks set in componentWillReceiveProps or boundary callback


    var callbacks = this._pendingCallbacks;

    if (callbacks) {
      this._pendingCallbacks = null;
      Updater.runCallbacks(callbacks, instance);
    }
  }
  /**
   * Call the component's `render` method and update the DOM accordingly.
   */
  ;

  _proto._updateRenderedComponent = function _updateRenderedComponent(context) {

    var prevRenderedComponent = this._renderedComponent;
    var prevRenderedElement = prevRenderedComponent._currentElement;
    var instance = this._instance;
    var nextRenderedElement;
    Host.component = this;
    performInSandbox(function () {
      {
        nextRenderedElement = instance.render();
      }
    }, instance);
    Host.component = null;

    if (shouldUpdateComponent(prevRenderedElement, nextRenderedElement)) {
      var prevRenderedUnmaskedContext = prevRenderedComponent._context;

      var nextRenderedUnmaskedContext = this._processChildContext(context);

      if (prevRenderedElement !== nextRenderedElement || prevRenderedUnmaskedContext !== nextRenderedUnmaskedContext) {
        prevRenderedComponent.updateComponent(prevRenderedElement, nextRenderedElement, prevRenderedUnmaskedContext, nextRenderedUnmaskedContext);
      }
    } else {
      var oldChild = prevRenderedComponent.getNativeNode();
      prevRenderedComponent.unmountComponent(true);
      this._renderedComponent = instantiateComponent(nextRenderedElement);

      this._renderedComponent.mountComponent(this._parent, instance, this._processChildContext(context), function (newChild, parent) {
        // TODO: Duplicate code in native component file
        if (!Array.isArray(newChild)) {
          newChild = [newChild];
        } // oldChild or newChild all maybe fragment


        if (!Array.isArray(oldChild)) {
          oldChild = [oldChild];
        }

        var driver = Host.driver; // If newChild count large then oldChild

        var lastNewChild;

        for (var i = 0; i < newChild.length; i++) {
          var child = newChild[i];

          if (oldChild[i]) {
            driver.replaceChild(child, oldChild[i]);
          } else if (lastNewChild) {
            driver.insertAfter(child, lastNewChild);
          } else {
            driver.appendChild(child, parent);
          }

          lastNewChild = child;
        } // If newChild count less then oldChild


        if (newChild.length < oldChild.length) {
          for (var _i = newChild.length; _i < oldChild.length; _i++) {
            driver.removeChild(oldChild[_i]);
          }
        }
      });
    }
  };

  _proto.getNativeNode = function getNativeNode() {
    var renderedComponent = this._renderedComponent;

    if (renderedComponent) {
      return renderedComponent.getNativeNode();
    }
  };

  _proto.getPublicInstance = function getPublicInstance() {
    var instance = this._instance; // The Stateless components cannot be given refs

    if (instance instanceof ReactiveComponent) {
      return null;
    }

    return instance;
  };

  return CompositeComponent;
}();

/**
 * Fragment Component
 */

var FragmentComponent =
/*#__PURE__*/
function (_NativeComponent) {
  _inheritsLoose(FragmentComponent, _NativeComponent);

  function FragmentComponent(element) {
    return _NativeComponent.call(this, element) || this;
  }

  var _proto = FragmentComponent.prototype;

  _proto.mountComponent = function mountComponent(parent, parentInstance, context, childMounter) {
    // Parent native element
    this._parent = parent;
    this._parentInstance = parentInstance;
    this._context = context;
    this._mountID = Host.mountID++;
    var instance = {
      _internal: this
    };
    this._instance = instance;
    var fragment = this.getNativeNode();
    var children = this._currentElement; // Process children

    this.mountChildren(children, context);

    if (childMounter) {
      childMounter(fragment, parent);
    } else {
      for (var i = 0; i < fragment.length; i++) {
        var child = fragment[i];
        Host.driver.appendChild(child, parent);
      }
    }

    return instance;
  };

  _proto.mountChildren = function mountChildren(children, context) {
    var _this = this;

    var renderedChildren = this._renderedChildren = {};
    var fragment = this.getNativeNode();
    var renderedChildrenImage = children.map(function (element, index) {
      var renderedChild = instantiateComponent(element);
      var name = getElementKeyName(renderedChildren, element, index);
      renderedChildren[name] = renderedChild;
      renderedChild._mountIndex = index; // Mount

      var mountImage = renderedChild.mountComponent(_this._parent, _this._instance, context, function (nativeNode) {
        if (Array.isArray(nativeNode)) {
          for (var i = 0; i < nativeNode.length; i++) {
            fragment.push(nativeNode[i]);
          }
        } else {
          fragment.push(nativeNode);
        }
      });
      return mountImage;
    });
    return renderedChildrenImage;
  };

  _proto.unmountComponent = function unmountComponent(notRemoveChild) {
    if (this._nativeNode) {
      Instance.remove(this._nativeNode);

      if (!notRemoveChild) {
        for (var i = 0; i < this._nativeNode.length; i++) {
          Host.driver.removeChild(this._nativeNode[i]);
        }
      }
    } // Do not need remove child when their parent is removed


    this.unmountChildren(true);
    this._currentElement = null;
    this._nativeNode = null;
    this._parent = null;
    this._parentInstance = null;
    this._context = null;
    this._instance = null;
  };

  _proto.updateComponent = function updateComponent(prevElement, nextElement, prevContext, nextContext) {
    // Replace current element
    this._currentElement = nextElement;
    this.updateChildren(this._currentElement, nextContext);
  };

  _proto.getNativeNode = function getNativeNode() {
    if (this._nativeNode == null) {
      this._nativeNode = [];
    }

    return this._nativeNode;
  };

  _proto.getPublicInstance = function getPublicInstance() {
    return this.getNativeNode();
  };

  _proto.getName = function getName() {
    return 'fragment';
  };

  return FragmentComponent;
}(NativeComponent);

function inject(_ref) {
  var driver = _ref.driver,
      measurer = _ref.measurer;
  // Inject component class
  Host.Empty = EmptyComponent;
  Host.Native = NativeComponent;
  Host.Text = TextComponent;
  Host.Fragment = FragmentComponent;
  Host.Composite = CompositeComponent; // Inject render driver

  Host.driver = driver || Host.driver;

  if (!Host.driver) {
    throw Error('Driver not found.');
  }
}

function render(element, container, options, callback) {
  // Compatible with `render(element, container, callback)`
  if (typeof options === 'function') {
    callback = options;
    options = null;
  }

  options = options || {}; // Init inject

  inject(options);
  var rootComponent = Instance.mount(element, container, options.parent);
  var componentInstance = rootComponent.getPublicInstance();

  if (callback) {
    callback.call(componentInstance);
  }

  return componentInstance;
}

var version = '1.0.0-beta';

var shared = {
  Host: Host,
  Instance: Instance,
  Element: Element,
  flattenChildren: flattenChildren
};

export { createElement, createRef, createContext, useState, useContext, useEffect, useLayoutEffect, useRef, useCallback, useMemo, useReducer, useImperativeHandle, Fragment, memo, render, version, shared };
//# sourceMappingURL=rax.mjs.map
